/***************************************************************************
 *            audio_win32.c
 *
 *  Fri Aug  6 15:17:58 2004
 *  Copyright  2004  benjie
 ****************************************************************************/

#include <gtk/gtk.h>
#include <stdio.h>
#include <windows.h>
#include <mmsystem.h>

#include "support.h"
#include "tocdoc.h"
#include "timeline.h"
#include "audio.h"
#include "audio_win32.h"

#define BLOCK_SIZE  8192
#define BLOCK_COUNT 20

// Win32 Specific Declarations
HWAVEOUT hWaveOut; /* device handle */
static CRITICAL_SECTION waveCriticalSection;
static WAVEHDR* waveBlocks;
static volatile int waveFreeBlockCount;
static int waveCurrentBlock;


// OSS Specific Audio Functions
static void CALLBACK waveOutProc(HWAVEOUT, UINT, DWORD, DWORD, DWORD);
static WAVEHDR* allocateBlocks(int size, int count);
static void freeBlocks(WAVEHDR* blockArray);


gboolean win32_startAudio(void)
{
    WAVEFORMATEX wfx; /* look this up in your documentation */
    /*
     * first we need to set up the WAVEFORMATEX structure. 
     * the structure describes the format of the audio.
     */
    wfx.nSamplesPerSec = 44100; /* sample rate */
    wfx.wBitsPerSample = 16; /* sample size */
    wfx.nChannels = 2; /* channels*/
    /*
     * WAVEFORMATEX also has other fields which need filling.
     * as long as the three fields above are filled this should
     * work for any PCM (pulse code modulation) format.
     */
    wfx.cbSize = 0; /* size of _extra_ info */
    wfx.wFormatTag = WAVE_FORMAT_PCM;
    wfx.nBlockAlign = (wfx.wBitsPerSample >> 3) * wfx.nChannels;
    wfx.nAvgBytesPerSec = wfx.nBlockAlign * wfx.nSamplesPerSec;
    /*
     * try to open the default wave device. WAVE_MAPPER is
     * a constant defined in mmsystem.h, it always points to the
     * default wave device on the system (some people have 2 or
     * more sound cards).
     */
    if(waveOutOpen(&hWaveOut, 
                   WAVE_MAPPER, 
                   &wfx, 
                   (DWORD_PTR)waveOutProc, 
                   (DWORD_PTR)&waveFreeBlockCount, 
                   CALLBACK_FUNCTION) != MMSYSERR_NOERROR) {
        fprintf(stderr, "unable to open WAVE_MAPPER device\n");
        return 0;
    }

    waveBlocks = allocateBlocks(BLOCK_SIZE, BLOCK_COUNT);
    waveFreeBlockCount = BLOCK_COUNT;
    waveCurrentBlock = 0;
    InitializeCriticalSection(&waveCriticalSection);

    return 1;
}


glong win32_writeBuffer(gchar *buffer, glong bytes)
{
    WAVEHDR* current;
    glong remain;
    glong bytesLeft = bytes;
    current = &waveBlocks[waveCurrentBlock];

    while(bytesLeft > 0) {
        /* first make sure the header we're going to use is unprepared */
        if(current->dwFlags & WHDR_PREPARED)
            waveOutUnprepareHeader(hWaveOut, current, sizeof(WAVEHDR));

        if(bytesLeft < (int)(BLOCK_SIZE - current->dwUser)) {
            memcpy(current->lpData + current->dwUser, buffer, bytesLeft);
            current->dwUser += bytesLeft;
            break;
        }

        remain = BLOCK_SIZE - current->dwUser;
        memcpy(current->lpData + current->dwUser, buffer, remain);
        bytesLeft -= remain;
        buffer += remain;
        current->dwBufferLength = BLOCK_SIZE;
        waveOutPrepareHeader(hWaveOut, current, sizeof(WAVEHDR));
        waveOutWrite(hWaveOut, current, sizeof(WAVEHDR));
        EnterCriticalSection(&waveCriticalSection);
        waveFreeBlockCount--;
        LeaveCriticalSection(&waveCriticalSection);

        /* wait for a block to become free */
        while(!waveFreeBlockCount)
            ms_sleep(10);

        /* point to the next block */
        waveCurrentBlock++;
        waveCurrentBlock %= BLOCK_COUNT;
        current = &waveBlocks[waveCurrentBlock];
        current->dwUser = 0;
    }
    return bytes;
}


gboolean win32_syncAudio(void)
{
    /* wait for all blocks to complete */
    while(waveFreeBlockCount < BLOCK_COUNT)
        ms_sleep(10);

    return 1;
}


gboolean win32_resetAudio(void)
{
    gint i;

    waveOutReset(hWaveOut);

    EnterCriticalSection(&waveCriticalSection);
    waveFreeBlockCount = BLOCK_COUNT;
    LeaveCriticalSection(&waveCriticalSection);

    waveCurrentBlock = 0;

    for(i = 0; i < BLOCK_COUNT; i++) {
        waveBlocks[i].dwUser = 0;
    }
    
    return 1;
}


void win32_closeAudio(void)
{
    gint i;

    /* unprepare any blocks that are still prepared */
    for(i = 0; i < waveFreeBlockCount; i++) 
        if(waveBlocks[i].dwFlags & WHDR_PREPARED)
            waveOutUnprepareHeader(hWaveOut, &waveBlocks[i], sizeof(WAVEHDR));

    DeleteCriticalSection(&waveCriticalSection);
    freeBlocks(waveBlocks);
    waveOutClose(hWaveOut);
}


static void CALLBACK waveOutProc(HWAVEOUT hWaveOut, 
				 UINT uMsg, 
				 DWORD dwInstance, 
				 DWORD dwParam1,
				 DWORD dwParam2)
{
    int* freeBlockCounter = (int*)dwInstance;
    
    /* ignore calls that occur due to openining and closing the device.*/
    if(uMsg != WOM_DONE)
        return;
    
    EnterCriticalSection(&waveCriticalSection);
    (*freeBlockCounter)++;
    LeaveCriticalSection(&waveCriticalSection);
}


WAVEHDR* allocateBlocks(int size, int count)
{
    unsigned char* buffer;
    int i;
    WAVEHDR* blocks;
    DWORD totalBufferSize = (size + sizeof(WAVEHDR)) * count;

    /* allocate memory for the entire set in one go */
    if((buffer = HeapAlloc(GetProcessHeap(), 
                           HEAP_ZERO_MEMORY, 
                           totalBufferSize)) == NULL) {
        fprintf(stderr, "Memory allocation error\n");
        ExitProcess(1);
    }

    /* and set up the pointers to each bit */
    blocks = (WAVEHDR*)buffer;
    buffer += sizeof(WAVEHDR) * count;

    for(i = 0; i < count; i++) {
        blocks[i].dwBufferLength = size;
        blocks[i].lpData = buffer;
        blocks[i].dwUser = 0;
        buffer += size;
    }

    return blocks;
}


void freeBlocks(WAVEHDR* blockArray)


{
    HeapFree(GetProcessHeap(), 0, blockArray);
}
