/***************************************************************************
 *            thumbnail.c
 *
 *  Thu Feb 10 18:23:38 2005
 *  Copyright  2005  User
 *  Email
 ****************************************************************************/

#include <gtk/gtk.h>
#include <stdio.h>
#include <stdlib.h>
#include <sndfile.h>

#include "tocdoc.h"
#include "timeline.h"
#include "audio.h"
#include "thumbnail.h"

GThread     *thumbThread;
GMutex      *thumbMutex = NULL;
GList       *thumbQueue = NULL;
gboolean    redrawWhenIdle;

extern tlType *tl;

thumbType *findThumbnailFor(gchar *filename);
void *makeAudioThumbnail(void *dummy);
gint thumbUpdate(gpointer data);
gint thumbIdler(gpointer data);



void createThumbnail(tocEntry *entry)
{
    thumbType *thumb;
    
    if (!thumbMutex) thumbMutex = g_mutex_new();

    thumb = findThumbnailFor(entry->filename);
    if (thumb) {
        g_mutex_lock(thumbMutex);
        thumb->useCount++;
        entry->thumbnail = thumb;
        g_mutex_unlock(thumbMutex);
        return;
    }

    entry->thumbnail = (thumbType *)g_malloc(sizeof(thumbType));
    entry->thumbnail->status = thumbNone;
    entry->thumbnail->useCount = 1;

    if (!thumbQueue) {
        thumbQueue = g_list_append(thumbQueue, entry);
        thumbThread = g_thread_create(&makeAudioThumbnail, NULL, TRUE, NULL);
        gtk_timeout_add(500, thumbUpdate, NULL);
        gtk_idle_add(thumbIdler, NULL);
    } else {
        g_mutex_lock(thumbMutex);
        thumbQueue = g_list_append(thumbQueue, entry);
        g_mutex_unlock(thumbMutex);
    }
}


void freeThumbnail(tocEntry *entry)
{
    if (!entry->thumbnail) return;
    
    entry->thumbnail->useCount--;
    if(entry->thumbnail->useCount > 0)
        return;
    
    if (thumbMutex) g_mutex_lock(thumbMutex);
    if (entry->thumbnail->status == thumbWorking)
        entry->thumbnail->status = thumbCancel;
    else
        thumbQueue = g_list_remove(thumbQueue, entry);
    if (thumbMutex) g_mutex_unlock(thumbMutex);

    while (entry->thumbnail->status == thumbCancel || entry->thumbnail->status == thumbWorking)
        ms_sleep(1);
    
    if (entry->thumbnail->data) {
        g_free(entry->thumbnail->data);
        entry->thumbnail->data = NULL;
    }
    g_free(entry->thumbnail);
    entry->thumbnail = NULL;
}


thumbType *findThumbnailFor(gchar *filename)
{
    GList *item;
    tocEntry *entry;

    if (!tl || !tl->document)
        return NULL;
    
    item = tl->document->entries;
	while (item) {
        entry = (tocEntry *)item->data;
        if (g_ascii_strcasecmp(entry->filename, filename) == 0)
            return entry->thumbnail;
        item = g_list_next(item);
    }
    
    return NULL;
}


void *makeAudioThumbnail(void *dummy)
{
    tocEntry    *entry;
    SNDFILE     *sndFile;
    SF_INFO     sfinfo;
    gshort      buffer[8 * 4];
    glong       size;
    glong       lenR;
    
    while (TRUE) {
        g_mutex_lock(thumbMutex);
        if (!thumbQueue) {
            g_mutex_unlock(thumbMutex);
            break;
        }
        
        entry = thumbQueue->data;
        if (!entry || !entry->filename) {
            thumbQueue = g_list_remove(thumbQueue, entry);
            g_mutex_unlock(thumbMutex);
            freeThumbnail(entry);
            continue;
        }
        
        sndFile = sf_open_relative(entry->filename, &sfinfo);
        if (!sndFile) {
            thumbQueue = g_list_remove(thumbQueue, entry);
            g_mutex_unlock(thumbMutex);
            freeThumbnail(entry);
            continue;
        }
        if (sfinfo.channels != 2 || !(sfinfo.channels & SF_FORMAT_PCM_16) ||
            sfinfo.samplerate != SAMPLES_PER_SEC) {
                thumbQueue = g_list_remove(thumbQueue, entry);
                printf("wrong soundfile type\n");
                sf_close(sndFile);
                g_mutex_unlock(thumbMutex);
                freeThumbnail(entry);
                continue;
        }
        size = sfinfo.frames / 8 * 2;
        entry->thumbnail->data = g_malloc(size);
        entry->thumbnail->length = 0;
    
        if (!entry->thumbnail->data) {
            sf_close(sndFile);
            thumbQueue = g_list_remove(thumbQueue, entry);
            g_mutex_unlock(thumbMutex);
            freeThumbnail(entry);
            continue;
        }
        entry->thumbnail->status = thumbWorking;
        g_mutex_unlock(thumbMutex);
        
        lenR=1;
        while (lenR) {
            g_mutex_lock(thumbMutex);
            if (entry->thumbnail->status == thumbCancel) {
                thumbQueue = g_list_remove(thumbQueue, entry);
                g_mutex_unlock(thumbMutex);
                break;
            }
            lenR = sf_read_raw(sndFile, buffer, 8*4);
            if (entry->thumbnail->length < size) {
                entry->thumbnail->data[(entry->thumbnail->length)++] = buffer[0]/256 + 128;
                entry->thumbnail->data[(entry->thumbnail->length)++] = buffer[1]/256 + 128;
            } else {
                (entry->thumbnail->length)++;
            }
            g_mutex_unlock(thumbMutex);
            if (entry->thumbnail->length % 4096 == 0) ms_sleep(1);
        }
        sf_close(sndFile);

        g_mutex_lock(thumbMutex);
        if (entry->thumbnail->status == thumbCancel) {
            thumbQueue = g_list_remove(thumbQueue, entry);
            entry->thumbnail->status=thumbDone;
            g_mutex_unlock(thumbMutex);
            continue;
        }
        thumbQueue = g_list_remove(thumbQueue, entry);
        entry->thumbnail->status = thumbDone;
        g_mutex_unlock(thumbMutex);
    }
    return 0;
}


gint thumbUpdate(gpointer data)
{
    redrawWhenIdle = TRUE;

    if (thumbQueue) {
        return TRUE;
    } else {
        return FALSE;
    }
}


gint thumbIdler(gpointer data)
{
    if (redrawWhenIdle) {
        redrawTimeline(tl);
        redrawWhenIdle = FALSE;
    }
    
    if (thumbQueue) {
        ms_sleep(10);
        return TRUE;
    } else {
        redrawTimeline(tl);
        return FALSE;
    }
}


void thumbnailGetRange(tocEntry *entry, glong start, glong end,
                       unsigned char *lmin, unsigned char *lmax,
                       unsigned char *rmin, unsigned char *rmax)
{
    glong increment;
    unsigned char sample;
    
    if (start == end) return;
    
    if (start < 0) start = 0;
    if (end > entry->thumbnail->length) end = entry->thumbnail->length;
    
    (*lmin)=255;
    (*lmax)=0;
    (*rmin)=255;
    (*rmax)=0;
    
    increment = MAX(2, 2*((end-start)/64));
    
    for (; start<end; start+=increment) {
        sample = entry->thumbnail->data[start];
        if (sample < (*lmin)) (*lmin) = sample;
        if (sample > (*lmax)) (*lmax) = sample;
        sample = entry->thumbnail->data[start+1];
        if (sample < (*rmin)) (*rmin) = sample;
        if (sample > (*rmax)) (*rmax) = sample;
    }
}
